package de.szut.lf8_starter.integration.project;

import de.szut.lf8_starter.project.ProjectEntity;
import de.szut.lf8_starter.project.ProjectRepository;
import org.json.JSONObject;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.*;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;

import java.time.LocalDate;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import static org.assertj.core.api.Assertions.assertThat;
import static org.hamcrest.Matchers.is;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@SpringBootTest
@AutoConfigureMockMvc(addFilters = false)
@DirtiesContext(classMode = DirtiesContext.ClassMode.AFTER_EACH_TEST_METHOD)
class CreateProjectActionTest {
    @Autowired
    private MockMvc mockMvc;
    @Autowired
    private ProjectRepository projectRepository;
    @Autowired
    private RestTemplate restTemplate;

    @Test
    void createProjectTest() throws Exception {
        String content = """
                {
                    "name": "name",
                    "leading_employee": 1,
                    "employees": [312],
                    "contractor": 4,
                    "contractor_name": "Peter File",
                    "comment": "goal of project",
                    "start_date": "01.01.2000",
                    "planned_end_date": "01.01.2001"
                  }
                """;

        final String contentAsString = this.mockMvc.perform(
                        post("/projects")
                                .content(content)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isCreated())
                .andExpect(jsonPath("name", is("name")))
                .andExpect(jsonPath("leading_employee", is(1)))
                .andExpect(jsonPath("employees", is(List.of(312))))
                .andExpect(jsonPath("contractor", is(4)))
                .andExpect(jsonPath("contractor_name", is("Peter File")))
                .andExpect(jsonPath("comment", is("goal of project")))
                .andExpect(jsonPath("start_date", is("01.01.2000")))
                .andExpect(jsonPath("planned_end_date", is("01.01.2001")))
                .andReturn()
                .getResponse()
                .getContentAsString();

        final long id = Long.parseLong(new JSONObject(contentAsString).get("id").toString());

        final Optional<ProjectEntity> project = this.projectRepository.findById(id);
        assertThat(project.get().getName()).isEqualTo("name");
        assertThat(project.get().getLeadingEmployee()).isEqualTo(1);
        assertThat(project.get().getContractor()).isEqualTo(4);
        assertThat(project.get().getContractorName()).isEqualTo("Peter File");
        assertThat(project.get().getComment()).isEqualTo("goal of project");
        assertThat(project.get().getStartDate()).isEqualTo(LocalDate.of(2000, 1, 1));
        assertThat(project.get().getPlannedEndDate()).isEqualTo(LocalDate.of(2001, 1, 1));
    }

    @Test
    void createProjectShouldReturnConflictResponseOnConflictingProjects() throws Exception {
        String content = """
                {
                    "name": "name",
                    "leading_employee": 1,
                    "employees": [312],
                    "contractor": 4,
                    "contractor_name": "Peter File",
                    "comment": "goal of project",
                    "start_date": "01.01.2000",
                    "planned_end_date": "01.01.2001"
                  }
                """;

        this.mockMvc.perform(
                        post("/projects")
                                .content(content)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isCreated());

        this.mockMvc.perform(
                        post("/projects")
                                .content(content)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isConflict());
    }

    @Test
    void createProjectWithPastStartDate() throws Exception {
        String content = """
                {
                    "name": "past project",
                    "leading_employee": 1,
                    "employees": [312],
                    "contractor": 4,
                    "contractor_name": "Past Contractor",
                    "comment": "past project",
                    "start_date": "01.01.1990",
                    "planned_end_date": "01.01.1991"
                  }
                """;

        this.mockMvc.perform(
                        post("/projects")
                                .content(content)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isCreated())
                .andExpect(jsonPath("name", is("past project")))
                .andExpect(jsonPath("start_date", is("01.01.1990")))
                .andExpect(jsonPath("planned_end_date", is("01.01.1991")));
    }

    @Test
    void createProjectWithFutureStartDate() throws Exception {
        String content = """
                {
                    "name": "future project",
                    "leading_employee": 1,
                    "employees": [312],
                    "contractor": 4,
                    "contractor_name": "Future Contractor",
                    "comment": "future project",
                    "start_date": "01.01.2100",
                    "planned_end_date": "01.01.2101"
                  }
                """;

        this.mockMvc.perform(
                        post("/projects")
                                .content(content)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isCreated())
                .andExpect(jsonPath("name", is("future project")))
                .andExpect(jsonPath("start_date", is("01.01.2100")))
                .andExpect(jsonPath("planned_end_date", is("01.01.2101")));
    }

    @Test
    void createProjectWithOverlappingDates() throws Exception {
        String content1 = """
                {
                    "name": "project 1",
                    "leading_employee": 1,
                    "employees": [312],
                    "contractor": 4,
                    "contractor_name": "Contractor 1",
                    "comment": "project 1",
                    "start_date": "01.01.2022",
                    "planned_end_date": "01.01.2023"
                  }
                """;

        String content2 = """
                {
                    "name": "project 2",
                    "leading_employee": 1,
                    "employees": [312],
                    "contractor": 4,
                    "contractor_name": "Contractor 2",
                    "comment": "project 2",
                    "start_date": "01.06.2022",
                    "planned_end_date": "01.06.2023"
                  }
                """;

        this.mockMvc.perform(
                        post("/projects")
                                .content(content1)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isCreated());

        this.mockMvc.perform(
                        post("/projects")
                                .content(content2)
                                .contentType(MediaType.APPLICATION_JSON)
                                .header(HttpHeaders.AUTHORIZATION, getBearerToken())
                )
                .andExpect(status().isConflict());
    }

    private String getBearerToken() {
        String url = "https://keycloak.szut.dev/auth/realms/szut/protocol/openid-connect/token";

        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);

        MultiValueMap<String, String> map = new LinkedMultiValueMap<>();
        map.add("grant_type", "password");
        map.add("client_id", "employee-management-service");
        map.add("username", "user");
        map.add("password", "test");

        HttpEntity<MultiValueMap<String, String>> request = new HttpEntity<>(map, headers);

        ResponseEntity<Map> response = this.restTemplate.exchange(url, HttpMethod.POST, request, Map.class);

        return Objects.requireNonNull(response.getBody()).get("access_token").toString();
    }
}

